Otis = LibStub("AceAddon-3.0"):NewAddon("Otis", "AceConsole-3.0", "AceEvent-3.0", "AceHook-3.0", "AceTimer-3.0")

--==== set upvalues ====--
-- GLOBALS: LibStub, BonusRollFrame, EncounterJournal, EncounterJournalEncounterFrameInfoLootTab, EncounterJournalEncounterFrameInfoModelTab, EncounterJournal_DisplayEncounter, EncounterJournal_DisplayInstance
local _G, pairs, rawset, select, type, CreateFrame, GetLootSpecialization, GetNumSpecializationsForClassID, GetSpecializationInfo, SetLootSpecialization, UnitClass, EJ_GetCreatureInfo, EJ_GetEncounterInfoByIndex, EJ_GetLootFilter = 
      _G, pairs, rawset, select, type, CreateFrame, GetLootSpecialization, GetNumSpecializationsForClassID, GetSpecializationInfo, SetLootSpecialization, UnitClass, EJ_GetCreatureInfo, EJ_GetEncounterInfoByIndex, EJ_GetLootFilter


Otis.DL = {
	[0] = "silent",
	"standard",
	"error",
	"test",
	"events",
	"config",
	"spam",
}
local DL = Otis.DL
do	-- create fast lookups in table for reverse lookup and index number as string
	for i=0, #DL do
		DL[ DL[i] ] = i
		DL[tostring(i)] = i
	end
end

local dbgprint = { print   = function(...) return Otis:Print(...) end,
                   squelch = function(...) return nil end }
local function SetDebugLevel(iDL)
	for i = 0,     iDL do dbgprint[ DL[i] ] = dbgprint.print;   end
	for i = iDL+1, #DL do dbgprint[ DL[i] ] = dbgprint.squelch; end
end

SetDebugLevel(DL.error)
-- SetDebugLevel(DL.config)


local function GetEncounters(t,k)
	local info = { ByIndex = {}, ByID = {} }

	local index = 1
	local name, _, encounterID = EJ_GetEncounterInfoByIndex(index)
	while encounterID do
		dbgprint.spam(index, name, encounterID)
		info.ByIndex[index] = encounterID
		info.ByID[encounterID] = index

		index = index + 1
		name, _, encounterID = EJ_GetEncounterInfoByIndex(index)
	end
	rawset(t, k, info)
	return info
end

local tblInstances = setmetatable( {}, { __index = GetEncounters } )
local tblSpecNames = { [0] = "Active" }
local tblSpecTex = {}
local btnLootSpec, btnUp, btnDown
local idPlayerClass
local dbbosses
local fWaitingOnBonusLoot = false
local idPendingSpec

local timerRestore = { func = "RestoreLootSpec", delay = 5 }
function timerRestore:Set(event)
	if self.timer then
		Otis:CancelTimer(self.timer)
	end
	self.timer = Otis:ScheduleTimer(self.func, self.delay)
	dbgprint.spam("scheduling restore for event", event)
end
function timerRestore:Cancel()
	if self.timer then
		Otis:CancelTimer(self.timer)
		self.timer = nil
	end
end


-- calling this with nil or 0 as the spec ID should hide all icons
local function ShowSpec(specID)
	for id, tex in pairs(tblSpecTex) do
		if id == specID then
			tex:Show()
		else
			tex:Hide()
		end
	end
end

function Otis:UpdateLootSpecButton()
	local instanceID, encounterID = EncounterJournal.instanceID, EncounterJournal.encounterID
	-- dbgprint.test("loot spec button update", instanceID, encounterID)
	if encounterID then
		ShowSpec(self.db.char.encounters[encounterID])
		btnLootSpec:Enable()
	else
		ShowSpec(nil)
		btnLootSpec:Disable()
	end
end
	
local function btnLootSpec_Click(self, button, down)
	local instanceID, encounterID = EncounterJournal.instanceID, EncounterJournal.encounterID

	if button == "LeftButton" then
		if encounterID then -- save the spec pref
			local classID, specID = EJ_GetLootFilter()
			
			if idPlayerClass ~= classID then
				dbgprint.error("ERROR: wrong class selected")
			elseif specID == 0 then
				dbgprint.error("ERROR: no spec selected")
			else
				-- change encounters table
				Otis.db.char.encounters[encounterID] = specID

				-- change bosses table
				local index = 1
				local id, name = EJ_GetCreatureInfo(index)
				while name do
					Otis.db.char.bosses[name] = specID
					index = index + 1
					id, name = EJ_GetCreatureInfo(index)
				end

				-- change button overlay icon
				ShowSpec(specID)
			end
		end
	elseif button == "RightButton" then
		if encounterID then -- clear setting for this encounter
			-- clear encounters table
			Otis.db.char.encounters[encounterID] = nil

			-- clear bosses table
			local index = 1
			local id, name = EJ_GetCreatureInfo(index)
			while name do
				Otis.db.char.bosses[name] = nil
				index = index + 1
				id, name = EJ_GetCreatureInfo(index)
			end

			-- hide button overlay icon
			ShowSpec(nil)
		end
	end
end

local function btnLootSpec_Enter(self)
	GameTooltip:SetOwner(self, "ANCHOR_RIGHT");
	GameTooltip:SetText("saved loot spec");
	GameTooltip:AddLine("left click to save spec for boss", 1, 1, 1, 1);
	GameTooltip:AddLine("right click to unset saved spec", 1, 1, 1, 1);
	GameTooltip:Show();
end

local function btnLootSpec_Leave(self)
	GameTooltip:Hide();
end

local function ScrollToBoss(index)
	local sf = _G.EncounterJournalEncounterFrameInfoBossesScrollFrame
	local btn = _G["EncounterJournalBossButton"..index]
	local diffBottom = sf:GetBottom() - btn:GetBottom()
	local diffTop = btn:GetTop() - sf:GetTop()

	if diffBottom > 0 then
		-- dbgprint.test(format("need to scroll the pane up by %i", diffBottom))
		sf:SetVerticalScroll(sf:GetVerticalScroll() + diffBottom)
	end
	if diffTop > 0 then
		-- dbgprint.test(format("need to scroll the pane down by %i", diffTop))
		sf:SetVerticalScroll(sf:GetVerticalScroll() - 10 - diffTop)
	end
end

local function btnUp_Click()
	if EncounterJournal then
		local instanceID, encounterID = EncounterJournal.instanceID, EncounterJournal.encounterID
		if instanceID and encounterID then
			local info = tblInstances[instanceID]
			local index = info.ByID[encounterID]
			if not index then
				dbgprint.test(format("btnUp_Click - nil index for encounter %i in instance %i", encounterID, instanceID))
			elseif index > 1 then
				EncounterJournal_DisplayEncounter(info.ByIndex[index-1], true)
				ScrollToBoss(index-1)
			elseif index == 1 then
				EncounterJournal_DisplayInstance(instanceID, true)
			end
		end
	end
end

local function btnDown_Click()
	if EncounterJournal then
		local instanceID, encounterID = EncounterJournal.instanceID, EncounterJournal.encounterID
		if instanceID then
			local info = tblInstances[instanceID]
			if encounterID then
				local index = info.ByID[encounterID]
				if not index then
					dbgprint.test(format("btnDown_Click - nil index for encounter %i in instance %i", encounterID, instanceID))
				elseif index < #info.ByIndex then
					EncounterJournal_DisplayEncounter(info.ByIndex[index+1], true)
					ScrollToBoss(index+1)
				end
			elseif #info.ByIndex > 0 then
				EncounterJournal_DisplayEncounter(info.ByIndex[1], true)
			end
		end
	end
end

function Otis:MakeButtons()
	if (type(EncounterJournalEncounterFrameInfoLootTab) ~= "table") then
		dbgprint.config("parent frame doesn't exist yet")
		return
	end

	dbgprint.config("making buttons")
	self:UnregisterEvent("EJ_LOOT_DATA_RECIEVED")
	self:UnregisterEvent("EJ_DIFFICULTY_UPDATE")

	btnLootSpec = CreateFrame("Button", "OtisButtonLootSpec", EncounterJournalEncounterFrameInfoModelTab)
	btnLootSpec:SetScript("OnClick", btnLootSpec_Click)
	btnLootSpec:RegisterForClicks("LeftButtonUp", "RightButtonUp")
	btnLootSpec:SetScript("OnEnter", btnLootSpec_Enter)
	btnLootSpec:SetScript("OnLeave", btnLootSpec_Leave)

	btnLootSpec:SetPoint("TOP", EncounterJournalEncounterFrameInfoModelTab, "BOTTOM", 0, 2)
	btnLootSpec:SetSize(63, 57)
	btnLootSpec:SetNormalTexture("Interface\\EncounterJournal\\UI-EncounterJournalTextures")
	btnLootSpec:GetNormalTexture():SetTexCoord(0.25585938, 0.37890625, 0.90332031, 0.95898438)	-- UI-EJ-Tab-Selected
	btnLootSpec:SetDisabledTexture("Interface\\EncounterJournal\\UI-EncounterJournalTextures")
	btnLootSpec:GetDisabledTexture():SetTexCoord(0.25585938, 0.37890625, 0.90332031, 0.95898438)	-- UI-EJ-Tab-UnSelected
	btnLootSpec:GetDisabledTexture():SetDesaturated(true)
	btnLootSpec:SetHighlightTexture("Interface\\EncounterJournal\\UI-EncounterJournalTextures")
	btnLootSpec:GetHighlightTexture():SetTexCoord(0.00195313, 0.12500000, 0.90332031, 0.95898438)	-- UI-EJ-Tab-Highlight

	for i=1, GetNumSpecializationsForClassID(idPlayerClass) do
		local id, name, _, icon = GetSpecializationInfo(i)
		local tex = btnLootSpec:CreateTexture(name, "OVERLAY")
		tex:SetTexture(icon)
		tex:SetPoint("BOTTOMLEFT", 20, 13)
		tex:SetPoint("TOPRIGHT", -11, -12)
		tex:Hide()
		tblSpecTex[id] = tex
	end

	btnLootSpec:Disable()

 	-- self:SecureHook("EncounterJournal_ListInstances", "UpdateLootSpecButton")
	self:SecureHook("EncounterJournal_DisplayInstance", "UpdateLootSpecButton")
	self:SecureHook("EncounterJournal_DisplayEncounter", "UpdateLootSpecButton")

	btnUp = CreateFrame("Button", "OtisButtonUp", EncounterJournalEncounterFrameInfoModelTab)
	btnUp:SetScript("OnClick", btnUp_Click)
--	btnUp:SetPoint("TOP", EncounterJournalEncounterFrameInfoModelTab, "BOTTOM", 6, 2)
	btnUp:SetPoint("TOP", btnLootSpec, "BOTTOM", 6, 2)
	btnUp:SetSize(63, 57)
	btnUp:SetNormalTexture("Interface\\ChatFrame\\UI-ChatIcon-ScrollUp-Up")
	btnUp:SetPushedTexture("Interface\\ChatFrame\\UI-ChatIcon-ScrollUp-Down")
	btnUp:SetDisabledTexture("Interface\\ChatFrame\\UI-ChatIcon-ScrollUp-Disabled")
	btnUp:SetHighlightTexture("Interface\\Buttons\\UI-Common-MouseHilight", "ADD")

	btnDown = CreateFrame("Button", "OtisButtonDown", EncounterJournalEncounterFrameInfoModelTab)
	btnDown:SetScript("OnClick", btnDown_Click)
	btnDown:SetPoint("TOP", btnUp, "BOTTOM", 0, 2)
	btnDown:SetSize(63, 57)
	btnDown:SetNormalTexture("Interface\\ChatFrame\\UI-ChatIcon-ScrollDown-Up")
	btnDown:SetPushedTexture("Interface\\ChatFrame\\UI-ChatIcon-ScrollDown-Down")
	btnDown:SetDisabledTexture("Interface\\ChatFrame\\UI-ChatIcon-ScrollDown-Disabled")
	btnDown:SetHighlightTexture("Interface\\Buttons\\UI-Common-MouseHilight", "ADD")


--[==[
	btnUp:SetNormalTexture("Interface\\MainMenuBar\\UI-MainMenu-ScrollUpButton-Up")
	btnUp:SetPushedTexture("Interface\\MainMenuBar\\UI-MainMenu-ScrollUpButton-Down")
	btnUp:SetDisabledTexture("Interface\\Buttons\\UI-ScrollBar-ScrollUpButton-Disabled")
	btnUp:SetHighlightTexture("Interface\\MainMenuBar\\UI-MainMenu-ScrollUpButton-Highlight", "ADD")
--]==]
end


function Otis:LoadSavedSettings()
	-- saved settings and default values
	local defaults = {
		global = {
			bossesSeen = {},
			output = 2,
		},
		char = {
			encounters = {},
			bosses = {},
		},
	}

	self.db = LibStub("AceDB-3.0"):New("OtisDB", defaults, true) --, _G.DEFAULT)

	-- we should be able to migrate saved data formats immediately
	local DBVERSION = 1
	
	if self.db.char.version == DBVERSION then
		return
	end
	
	if self.db.char.version == nil then
		self.db.char.version = DBVERSION
		return
	end

	self.db.char.version = DBVERSION
end


-- Called when the addon is enabled
function Otis:OnEnable()
	self:LoadSavedSettings()

	idPlayerClass = select(3, UnitClass("player"))
	for i=1, GetNumSpecializationsForClassID(idPlayerClass) do
		local id, name = GetSpecializationInfo(i)
		tblSpecNames[id] = name
	end

	-- TODO - investigate DC case as well as whether trying to RestoreLootSpec this early was causing dc's
	-- self:RestoreLootSpec("OnEnable")  -- in case you DC in combat

	self:RegisterEvent("PLAYER_LOOT_SPEC_UPDATED")

	self:RegisterEvent("PLAYER_REGEN_DISABLED")
	self:RegisterEvent("PLAYER_REGEN_ENABLED")

	self:RegisterEvent("PLAYER_DEAD")
	self:RegisterEvent("PLAYER_ALIVE")
	self:RegisterEvent("PLAYER_UNGHOST")
--	self:RegisterEvent("PLAYER_ENTERING_WORLD")

	self:SecureHookScript(BonusRollFrame, "OnShow", "WaitForBonusLoot")
	self:SecureHookScript(BonusRollFrame, "OnHide", "WaitFinished")
	self:RegisterEvent("MISSING_OUT_ON_LOOT")

	self:RegisterEvent("EJ_LOOT_DATA_RECIEVED", "MakeButtons")
	self:RegisterEvent("EJ_DIFFICULTY_UPDATE",  "MakeButtons")
end

-- Called when the addon is disabled
function Otis:OnDisable()
	self:UnregisterEvent("PLAYER_LOOT_SPEC_UPDATED")

	self:UnregisterEvent("PLAYER_REGEN_DISABLED")
	self:UnregisterEvent("PLAYER_REGEN_ENABLED")

	self:UnregisterEvent("PLAYER_DEAD")
	self:UnregisterEvent("PLAYER_ALIVE")
	self:UnregisterEvent("PLAYER_UNGHOST")
--	self:UnregisterEvent("PLAYER_ENTERING_WORLD")

	self:UnregisterEvent("EJ_LOOT_DATA_RECIEVED")
	self:UnregisterEvent("EJ_DIFFICULTY_UPDATE")

	self:UnregisterEvent("MISSING_OUT_ON_LOOT")

	self:UnhookAll()
end

function Otis:PLAYER_REGEN_DISABLED() -- entered combat
	dbgprint.events("PLAYER_REGEN_DISABLED - combat started")
	dbbosses = self.db.char.bosses

	if self.db.char.savedLootSpec then
		-- we could be reentering combat with boss after a resurrection or after a disconnect
		-- or someone could also have started a new fight while we're still deciding whether 
		-- to roll in the bonus loot dialog
		if fWaitingOnBonusLoot then
			-- if the bonus loot roll window really is still up then we really don't want to do 
			-- anything with it until the user is done. we should exit without setting subscribing
			-- to any events. we can retry things when the loot window closes
			if BonusRollFrame and BonusRollFrame:IsVisible() then
				dbgprint.events(format("loot window for %s still up; ignoring start of new combat until resolved", self.db.char.savedLootBoss))
				return
			else
				dbgprint.events(format("player presumed to have died fighting %s; do not wait on bonus roll window"))
				fWaitingOnBonusLoot = false
			end
		end

		dbgprint.events(format("player may have wiped or reentered combat with %s after resurrection or disconnect. restore, clear and rerun boss detection", self.db.char.savedLootBoss))
		self:RestoreLootSpec()
	end

	if IsInInstance() then
		self:RegisterEvent("INSTANCE_ENCOUNTER_ENGAGE_UNIT")
		
		-- frames can sometimes go up before we enter combat, 
		-- especially when stealthed. force a check just in case
		self:INSTANCE_ENCOUNTER_ENGAGE_UNIT("INSTANCE_ENCOUNTER_ENGAGE_UNIT - forced at combat start")
	else
--		self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
		self:RegisterEvent("PLAYER_TARGET_CHANGED")
		self:RegisterEvent("UPDATE_MOUSEOVER_UNIT")
		self:RegisterEvent("UNIT_CLASSIFICATION_CHANGED")

		-- They may never actually CHANGE target in the fight if they start
		-- out targeted on the boss. So, call the handler at the start.
		self:PLAYER_TARGET_CHANGED("PLAYER_TARGET_CHANGED - forced at combat start")
	end



end

function Otis:INSTANCE_ENCOUNTER_ENGAGE_UNIT(...)
	dbgprint.events(...)
	
	for i=1, 5 do
		local unit = "boss"..i
		local name = UnitName(unit)
		if name then
			self.db.global.bossesSeen[name] = true
			dbgprint.test(unit,name)
			local spec = dbbosses[name]
			if spec then
				self:ChangeLootSpec(spec, name, unit)

				-- we've identified this encounter, don't need to process more target updates
				self:UnregisterEvent("INSTANCE_ENCOUNTER_ENGAGE_UNIT")
				return
			end
		end
	end
end

function Otis:PLAYER_TARGET_CHANGED()
	self:CheckUnit("target")
end

function Otis:UPDATE_MOUSEOVER_UNIT()
	self:CheckUnit("mouseover")
end

function Otis:UNIT_CLASSIFICATION_CHANGED(event, unit)
	-- surely this is a rare event
	dbgprint.events(event, unit)
	self:CheckUnit(unit)
end

function Otis:CheckUnit(unit)
	if not UnitExists(unit) then return end
	
	if UnitIsPlayer(unit) then
		unit = unit .. "target"
		if not UnitExists(unit) then return end
	end

	-- the stupid celestials don't appear to be worldbosses ... how lame
	local classification = UnitClassification(unit)
	if  classification == "worldboss" or classification == "elite" then
		-- UnitIsTappedByPlayer(unit) -- at the moment going to say that we don't really care if the unit has been tapped by the player or not because we see no real harm
		local name = UnitName(unit)
		local spec = dbbosses[name]
		if spec then
			-- dbgprint.test("CheckUnit would have changed loot spec here for", unit, name)
			self:ChangeLootSpec(spec, name, unit)
		end

		-- we've identified this encounter, don't need to process more target updates
		self:UnregisterEvent("PLAYER_TARGET_CHANGED")
		self:UnregisterEvent("UPDATE_MOUSEOVER_UNIT")
		self:UnregisterEvent("UNIT_CLASSIFICATION_CHANGED")
	end
end

function Otis:COMBAT_LOG_EVENT_UNFILTERED(event, _, _, _, _, srcName, _, _, _, dstName)
	-- local specID = dbbosses[srcName] or dbbosses[dstName]
	local name, spec = srcName, dbbosses[srcName]
	if not spec then
		name, spec = dstName, dbbosses[dstName]
	end

	if spec then
		self:ChangeLootSpec(spec, name, "")
		-- we've identified this encounter, don't need to process more combat log notifications
		self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	end
end

function Otis:PLAYER_REGEN_ENABLED(event) -- left combat
	dbgprint.events("PLAYER_REGEN_ENABLED - combat ended")
	-- self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	self:UnregisterEvent("PLAYER_TARGET_CHANGED")
	self:UnregisterEvent("UPDATE_MOUSEOVER_UNIT")
	self:UnregisterEvent("UNIT_CLASSIFICATION_CHANGED")
	self:UnregisterEvent("INSTANCE_ENCOUNTER_ENGAGE_UNIT")
	
	if self.db.char.savedLootSpec and not fWaitingOnBonusLoot then
		timerRestore:Set(event)
	end
end

function Otis:PLAYER_LOOT_SPEC_UPDATED(event)
	local newspec = GetLootSpecialization()
	if not idPendingSpec then
		dbgprint.events(format("PLAYER_LOOT_SPEC_UPDATED to %s - user initiated", tblSpecNames[newspec]))
	elseif newspec == idPendingSpec then
		dbgprint.events(format("PLAYER_LOOT_SPEC_UPDATED to %s - expected by Otis", tblSpecNames[newspec]))
	else
		dbgprint.error(format("ERROR: PLAYER_LOOT_SPEC_UPDATED to %s, expected %s", tblSpecNames[newspec], tblSpecNames[idPendingSpec]))
	end
	idPendingSpec = nil	
end

function Otis:ChangeLootSpec(spec, name, unit)
	local current = GetLootSpecialization()
	local oldspec = idPendingSpec or current

	if idPendingSpec then
		if spec == idPendingSpec then
			dbgprint.events(format("a change to %s is already pending", tblSpecNames[spec]))
		else
			self.db.char.savedLootSpec = idPendingSpec
			self.db.char.savedLootBoss = name

			-- if spec == current we will not get a notification and should not set a new value for pending
			-- the no-op does prevent the pending for happening, though, so must be done
			if spec == current then
				dbgprint.events(format("reverting spec change from %s to %s for %s %s", tblSpecNames[spec], tblSpecNames[idPendingSpec], unit, name))
				SetLootSpecialization(spec)
				idPendingSpec = nil
			else
				dbgprint.events(format("changing loot spec from %s to %s for %s %s", tblSpecNames[idPendingSpec], tblSpecNames[spec], unit, name))
				SetLootSpecialization(spec)
				idPendingSpec = spec
			end
		end
	elseif current ~= spec then
		self.db.char.savedLootSpec = current
		self.db.char.savedLootBoss = name

		dbgprint.events(format("changing loot spec from %s to %s for %s %s", tblSpecNames[current], tblSpecNames[spec], unit, name))
		SetLootSpecialization(spec)
		idPendingSpec = spec
	end
end

function Otis:RestoreLootSpec()
	if fWaitingOnBonusLoot then
		dbgprint.test("ERROR - restore called while we're supposed to be waiting on bonus loot")
		fWaitingOnBonusLoot = false
	end

	if self.db.char.savedLootSpec then
		dbgprint.events(format("restoring loot spec to %s after %s fight", tblSpecNames[self.db.char.savedLootSpec], self.db.char.savedLootBoss))
		idPendingSpec = self.db.char.savedLootSpec
		SetLootSpecialization(self.db.char.savedLootSpec)
		self.db.char.savedLootSpec = nil
		self.db.char.savedLootBoss = nil
	end
end

function Otis:WaitForBonusLoot()
	-- if we didn't change spec, then we don't have to wait. this boss means nothing to us
	if self.db.char.savedLootSpec then
		dbgprint.events("waiting for bonus loot")

		timerRestore:Cancel()
		fWaitingOnBonusLoot = true
	end
end

function Otis:WaitFinished()
	fWaitingOnBonusLoot = false
	if InCombatLockdown() then
		-- if we enter combat while bonus loot window is still up, we don't
		-- try to identifier the encounter or do anything else. now that the
		-- bonus loot has been dealt with, start over with the current fight
		self:PLAYER_REGEN_DISABLED()
	else
		timerRestore:Set("BonusRollFrame_OnHide")
	end
end

function Otis:MISSING_OUT_ON_LOOT(event)
	dbgprint.test(event)
	fWaitingOnBonusLoot = false
	timerRestore:Set(event)
end

function Otis:PLAYER_DEAD(event)
	dbgprint.events(event)
	self:WaitForBonusLoot()
end

function Otis:PLAYER_ALIVE(event)
	dbgprint.events(event)
	-- self:WaitFinished()
end

function Otis:PLAYER_UNGHOST(event)
	dbgprint.events(event)
	-- self:WaitFinished()
end

function Otis:state()
	local current = tblSpecNames[GetLootSpecialization()]
	local savedBoss = self.db.char.savedLootBoss or "nil"
	local savedSpec = self.db.char.savedLootSpec and tblSpecNames[self.db.char.savedLootSpec] or "nil"
	local pending = idPendingSpec and tblSpecNames[idPendingSpec] or "nil"
	dbgprint.standard(format("current: %s  savedBoss: %s  savedSpec: %s  waiting: %s  pending: %s", current, savedBoss, savedSpec, tostring(fWaitingOnBonusLoot), pending))
end